"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                         BIT TWIDDLING (AKA BIT FLIPPING, BIT MANIPULATION)                  ||
      ||        REFERENCES: 1. Hacker's delight by Henry S. Warren, Jr. Addison-Wesley, 2003.        ||
      +=================-=========-=========-=========*=========-=========-=========-=================+

                        Programmers have gradually discovered that bitwise operations
                        can be amazingly useful.--Donald E. Knuth, The Art of Computer
                        Programming Vol. 4, Part 1 (Addison Wesley, 2009), page 2.


   Bit flipping in this discussion is done with POSITIVE INTEGERS only. The input and output of all integers
defaults to base ten. So, we tend to think the internal representations and computations are in base ten
(denary), but they are in base two (binary). Sometimes we can achieve impressive gains in memory and speed if
we make the effort to work with the bits. In many cases binary arithmetic is simple to understand.

   Note well: ALL integers print out in base ten, unless you print with the bin() function--e.g.

                        print (bin(5))                  # output: 0b101
                        print (bin(5)[2:])              # output: 101
                        print (type(bin(5)))            # output: <class 'str'>
                        print (isinstance(bin(5), str)) # output: True
                        print (int('101', 2))           # output: 5

    Python provides special operators and functions to work with bits.

    +--------- PYTHON BIT OPERATORS, FUNCTIONS, AND TRICKS ----------------+
    |  operator          meaning                  examples                 |
    |   &          and                     1010 & 1100 = 1000              |
    |   |          non-exclusive or        1010 | 1100 = 1110              |
    |   ^          exclusive or            1010 ^ 1100 = 0100              |
    |   ~          not (flip all bits)    ~1010        = ????              |
    |              flip all bits           0x1F-(0b11010) = 101            |
    |   <<         shift left  n bits      101    << 3 = 101000            |
    |   >>         shift right n bits      101000 >> 3 = 101               |
    +----------------------------------------------------------------------+
    |   0b or 0B   interpret as binary      0b10101    = 21                |
    |   bin()      express as binary        bin(21)    = 10101             |
    +----------------------------------------------------------------------+
    |  turn ON  nth bit from right: num |=  (1 << n)                       |
    |  turn OFF nth bit from right: num &= ~(1 << n)                       |
    |  flip     nth bit from right: num ^=  (1 << n)                       |
    |  test     nth bit from right: if (num & (1 << n))  > 0: ...          |
    |                               if (num & (1 << n)) == 0: ...          |
    |  clear the right-most bit:    num = num & (num-1)                    |
    |  smear right-most 1 to right: num | (num-1)                          |
    |  extract right-most 1:        num = num & -num (e.g., 101100 ->10)   |
    |  extract nth bit from left:   bit = (num >> n) & 1                   |
    |  mod 2**n:                    x mod 2**n = x &(2**n - 1)             |
    +----------------------------------------------------------------------+
    |  Below, the 0 is a zero, not a letter in 0b (= 0B).                  |
    |  print (0b10101)           # = 21                                    |
    |  print ( int("10101", 2) ) # = 21 (string to binary integer)         |
    |  print(bin(21))            # = 0b10101                               |
    |  print(bin(21)[2:])        # =   10101                               |
    |  n = 0b1001                                                          |
    |  print(n.bit_length())     # = 4                                     |
    +----------------------------------------------------------------------+

   Notes:
    1. A power of 2 has only a single 1. Thus, 2**5 = b100000, and 2**5-1 = b11111.
    2. All even numbers have a zero in the right-most bit.
    3. The binary b101 0101 1000 1010 equals 0x558A in hexadecimal (group by fours). You must see
       this as obvious.
    4. 32768 = (1 << 15) = 0x8000 = pow(2, 15) = 2**15.

   It is possible to work with the bits in a list of integers: Let x have 96 bits, contained in a
list of three 32-bit integers.

                x = [2**32-1, 2**32-1, 2**32-1]
                n = 90 # = the 90th bit
                x[n//32] &= ~(1<<(n %32)) # turn OFF bit n
                x[n//32] |=  (1<<(n %32)) # turn ON  bit n

   Perhaps easier is just to work with a single LARGE string x = [2**97-1]. The extra 97th bit
is to always keep the string beginning with a 1, not a 0.
#-----------------------------------------------------------------------------------------

        The 1s complement of 10010110 is
                             01101001 (inversion of all digits).

        The 2s complement of 10010110 is
                             01101010 (inversion of all digits and add 1).
#-----------------------------------------------------------------------------------------

  Question: What is the 4-bit binary representation of -13 (= -0b1101)?
  Answer:   0011 (2s complement). Note 1101 + 0011 = 13 + (-13) = 0, always.
  Why:      This is how we can quickly find the binary representation for a negative
            binary number, and check our work by adding. You can, of course, do the re-
            verse to move from -13 to +13, by subtracting 1, and then taking the 1s
            complement of the remainder.

#-----------------------------------------------------------------------------------------
   Below is a scheme to use 19 bits to represent a tic-tac-toe board. We need 2 bits for
each of the 9 cells. But we cannot allow the binary number to begin with a zero. So, we
append a fixed digit of 1 in the left-most place. Let 0b00 mean empty cell; let 0b10 mean
'X' and let 0b01 mean 'O'. The value 0b11 is not used.

     X|O|X
    --+-+--         O  X  O  O  X  O  X  O  X
     O|X|O   = 0b1 01 10 01 01 10 01 10 01 10  <-- 19 bits
    --+-+--
     O|X|O

   Below is one way to set up the board. Note that 0b10 = 2 and 0b01 = 1.

    board = 0b1000000000000000000 # = empty board (only 19 bits!)
    board |= (0b10)               # = X in top left corner
    board |= (0b01 <<  2)         # = O in top MIDDLE
    board |= (0b01 <<  4)         # = O in top right corner
    board |= (0b01 << 16)         # = O in bottom right corner
#-----------------------------------------------------------------------------------------

   The number of 1-bits in a binary number x is called its POPULATION COUNT. How do we
find it? Better, how do we find it so that our code is easy-to-understand. Unless the
population count is at the center of a loop in a slow program, we are not much interested
in speed. Anyway, here is my solution: bin(x)[2:].count('1').

    x = 851 # = 1101010011 in binary with pop. count = 6
    print(bin(x))                 # = 0b1101010011, a string type
    print(bin(x)[2:])             # =   1101010011, a string type
    print (bin(x)[2:].count('1')) # = 6, an int type.

   The HAMMING DISTANCE between two numbers is the count of binary differences between the
two numbers. The Hamming distance between 10010 and 11100 is 3. Does it make a difference
to compare two number of a different size? No, because there is no such thing as two
binary numbers of a different size. All binary numbers and even all numbers have an in-
finite number of zeroes attached to their left side. All numbers are countably infinitely
long, and therefore the same size. Anyway, our answer is bin(x^y)[2:].count('1')

    x =  1500 # = 01101010011
    y =  456 # = 11000110111
    print('x =  ',   bin(x)[2:])           # x =   10111011100
    print('y =    ', bin(y)[2:])           # y =     111001000
    print('x^y =',   bin(x^y)[2:])         # x^y = 10000010100
    print('1s =', bin(x^y)[2:].count('1')) # 1s = 3
#-----------------------------------------------------------------------------------------

   How do you create a number of all 1s? The following expression is a binary number made
up of exactly max 1s: int('1' * max, 2).

   How can you use bits to create a list of prime numbers? I give that below, and then
re-write the method to work with bits, thereby saving much memory.

#----------------------=---------------------=---------------------=------------
# Create a list of max ints. Initially set every int to 1. By the sieve method of
# Eratosthenes, set to zero any int whose position number is not a prime number.
#
def sieveOfEratosthenesUsingLists(max): # max = the largest number in the list
    p = [1 for n in range(max)]
    for d in range (2, int(sqrt(max)+1)):
        for n in range (d+1,max):
            if n % d == 0:
                p[n] = 0
    print('PRIME NUMBERS:', end= '')
    for n in range (2, max):
        if p[n] == 1: print (n, end = ', ') # output:2, 3, 5, 7, 11, 13, 17, 19, 23, etc.
    print()
#--------------------------------------------------------------------------------

# Create a binary number of max bits. Initially set every bit to 1. By the sieve method
# of Eratosthenes, set to zero any bit whose position number is not a prime number.
#
def sieveOfEratosthenesUsingBits(max):     # max = the number of bits
    p = int('1' * max, 2)                  # create a binary number with max 1s.
    for d in range (2, int(sqrt(max)+1)):  # create list of divisors
        for n in range (d+1,max):          # run through bits
            if n % d == 0:                 # if position is evenly divided it is not prime
                p &= ~(1 << (max-n-1))     # turn off nth bit from left

    print('PRIME NUMBERS:', end= '')
    for n in range (2, max):
        if (p & (1 << (max-n-1))) > 0:     # print position
            print (n, end = ', ') # output:2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, etc.

   Perhaps a better way to create a list of prime numbers is given below:

    p = [2,]
    for n in range (3, 200):
       for divisor in p:
           if n%divisor == 0:
              break
       if n%divisor != 0:
          p.append(n)
    print (p)
#---------------------------------------------------------------------------------------------------

How to pack two 8-bit bytes (low and high) into a 16-bit word:
    word = (high << 8)  + low
    word = (high << 8)  | low
    word = (high * 256) + low
#-------------------------------------------------------------------------------------------------------------

An example of searching for duplicates.
def hasDuplicates(numList):
    lookUp = {0:1, 1:2, 2:4, 3:8, 4:16, 5:32, 6:64, 7:128, 8:256, 9:512, 10:1024, 11:2048, 12: 4096, 13:8192,
              14:16384, }
    sum = 0                         # Turn all bits off.
    for value in numList:
        pwr = 2**value              # or use the look up table: pwr = lookUp[value]
        if (pwr & sum): return True # If the pwr bit is already on, then we have a duplicate.
        sum |= pwr                  # Turn the pwr bit on.
    return False
#-------------------------------------------------------------------------------------------------------------

def main():
    numList = [2,4,4]
    print (numList, 'has duplicates is', hasDuplicates(numList))

#-------------------------=-----------------------=-------------------------=-----------------------------
if __name__ == '__main__':  main()
###############################################<END OF PROGRAM>###############################################
