"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                        CLASSES AND OBJECT ORIENTED PROGRAMMING (OOP)                        ||
      ||                               (also overloading operators)                                  ||
      ||   REFERENCE: Steve McConnell, Code Complete, 2nd Ed. (Microsoft Press, 2004), pages 92-97.  ||
      +=================-=========-=========-=========*=========-=========-=========-=================+


   Classes allow the user to create new types of variables (instances), with built-in functions, with a simple
syntax. Classes are also instance factories. Below is a simple example of a python class.

class Fraction(object):
    quotient = 0                      # a class (aka static) variable is a single place in memory
                                      # that is accessed by all objects of the class type.
    def __init__(self, n = 0, d = 1): # the constructor
        self.num = n                  # an INSTANCE variable
        self.den = d                  # an INSTANCE variable
        quotient = n/d                # an INSTANCE variable

    def __repr__(self):               # the representation string
        return str(self.num) + '/' + str(self.den)

    def print(self):                  # making print a method
        print (self.num, '/', self.den, sep = '')

    def mult1(self, fract):
        return self.num*fract.num, self.den*fract.den

    def mult2(fract1, fract2): # Note that self is omitted.
        return fract1.num*fract2.num, fract1.den*fract2.den

def main():
    x = Fraction(2,3)
    x.quotient = 2/3
    x.print()                  # output: 2/3
    print(x)                   # output: 2/3
    print(x.quotient)          # output: 0.6666666666666666
    print(Fraction.quot)       # output: 0
    x.q = 100                  # creating a new instance variable.
    print (x.q * x.quotient)   # output: 66.66666666666666
    y = Fraction (4,5)
    print(x.mult1(y))          # output: (8, 15)
    print(y.mult1(x))          # output: (8, 15)
    print(Fraction.mult2(x,y)) # output: (8, 15) Note: This method can be called without an object.
if __name__ == '__main__'

WARNING: Note well the ability to append an instance variable to a particular object (x.q = 100, above). This
option is more of a curse than a benefit. First you should have built the variable in to the class in the
first place. Second, if you misspell an instance variable name, no error will be generated. Instead, the
computer thinks you have just created a new instance variable. This can be a difficult variable to track down.

   Instance variables are default when a reference is indeterminate between a class variable and an instance
variable of the same name. But why would anyone use the same name? Anyway, here is an example. Note that there
are two variables named a, below.

            class doit():
                a = 1 # class variable (aka class attribute)
                b = 3 # class variable
                def __init__(self):
                   self.a = 2 # instance variable (aka instance attribute)

            def main():
                x = doit()
                print(doit.a, x.a, x.b) # output 1,2,3
==============================================================================================================




   1. In classes, operators can be overloaded (polymorphism), unlike in Java. Unless you are building a mathe-
      matical class--e.g., vectors or fractions--you are unlikely to want to overload operators. However,
      Python does NOT allow polymorphism of functions or methods. In other words, you can NOT have two
      functions with the same name. Here is an example of the failure.

                def doit(a):
                    print(a)
                def doit(a,b):
                    print(a,b)
                def main():
                     doit(1) # ERROR: doit() missing 1 required positional argument: 'b'
                     doit(1,2)

   2. Curiously, the assignment operator (=) can NOT be overloaded in Python. Thus, you may need to write an
      assignment function equals(), so that A.equals(B) gives A a copy of the contents of B, but keeps A and B
      stored in separate memory locations.

   3. To copy an object you cannot use the assignment operator (=). To write V = W, means V and W are two
      different names for the same place in memory. To generate a true copy, you must use deepcopy() imported
      from the copy module:

                                        from copy import deepcopy
                                        Obj2 = deepcopy(Obj1)

   For small programs under 500 lines one might use an object as a bundle of data bound with its own pro-
cessing logic, but is unlikely to use inheritance. Inheritance is more of an academic topic than a practical
topic for people who write small programs (scripts).

   An object (aka an INSTANCE) is formed (INSTANTIATED) by a binding of data (instance variables and class
variables both called "attributes") with methods (functions called "behaviors"). Both data and methods are
bound into an object that is accessed through a single identifier, assisted by the dot operator. The design
template for a particular object is called its CLASS. This binding of data and methods is called
ENCAPSULATION. The positive features of using objects are

   1) INFORMATION HIDING--The programmer doesn't have to understand or recall how certain methods work to use
      those methods.

   2) DATA PROTECTION--There is some protection against accidentally modifying some data or entering bad data.
      The programmer is discouraged from changing certain lines of code. In Python, which cannot declare data
      to be private, use getters and setters.to keep the software trustworthy.
   3) INHERITANCE--You can easily create new objects based on trusted code.
   4) POLYMORPHISM--One operator (but not method) works with different kinds of data.
   5) EASE OF CODING--This is due the binding of data with its own processing logic (methods). Reducing pro-
      gram complexity is the most important reason to implement classes. The instance and class variables in a
      class are semi-global. They don't have to be passes or returned.

   Methods beginning and ending with two underscores (like __mul__ for the binary * operator) are called
SPECIAL FUNCTIONS, and used to overload operators. Below is a partial list. See the Internet for many more.

                __add__ (self, x, y)            x +  y
                __radd__(self, x, y)            y +  x
                __iadd__(self, x, y)            x += y

                __sub__ (self, x, y)            x -  y
                __rsub__(self, x, y)            y -  x
                __isub__(self, x, y)            x -= y

                __mul__ (self, x, y)            x *  y
                __rmul__(self, x, y)            y *  x
                __imul__(self, x, y)            x *= y

                __truediv__ (self, x, y)        x /  y
                __itruediv__(self, x, y)        x /= y
                __rtruediv__(self, x, y)        y /= x

                __floordiv__ (self, x, y)       x // y
                __rfloordiv__(self, x, y)       y // x

                __mod__(self, x, y)             x %  y
                __pow__(self, x, y)             x ** y

                __lt__(self, x, y)              x <  y
                __le__(self, x, y)              x <= y
                __eq__(self, x, y)              x == y
                __ne__(self, x, y)              x != y
                __gt__(self, x, y)              x >  y
                __ge__(self, x, y)              x >= y

                __getitem__(self, index)        self[index]
                __setitem__(self, index, num)   self[index] = num

   The following is an example in overloading the square brackets, when Obj is an object with a tuple of
instant variables. [Believe it or not, both __getitem__ and __setitem__ work for slicing as well as indexing.]

   def __init__(self, *data):          # __init__ assigns a tuple to data--e.g.,
       self.data = list(data)          # data = (0,1,2, ...)

   def __getitem__ (self, index):      # Overload square brackets
       return self.data[index]         # return Obj[index]

   def __setitem__ (self, index, num): # Overload square brackets
       self.data[index] = num          # Obj[index] = num
       return self
#-------------------------------------------------------------------------------------------------------------

                         ESSAY: Why Do We Need Classes (Object Oriented Programming)?
                                   by M. Stueben (August 20, 2014)


   The short answer is that we don't need to use classes in school problems. Classes can help in some cases,
but only a little. We first note that there is not a significant difference between the two calls X.doit() and
doit(X).

A claimed justification for classes is code reuse through inheritance. Of course, we already have code reuse
through cut-and-paste and through importing library files (modules). And some classes get so coupled with
their applications, that they are not easily reused. The advantages of code reuse through classes are more
appreciated in industry than in school problems.

   That being said, I once used inheritance to import four functions (methods) into a vector class. Those
functions were only applicable to the particular problem I was coding, and I did not want my vector class to
get modified. But that was more of a composition of two classes (a Vector class and a set of Nelder-Mead
functions/methods) than inheritance.

   Another advantage of inheritance is that a single change to a parent is a change to all of its children,
because the commonality of the children's code resides only in one place: the parent. But even for programs
without classes, commonality can be factored out into simple functions.

   The next advantage of classes is encapsulation. If the class models something in reality or even the
programmer's perspective on a problem, then he can think and write code in terms of objects instead of their
tiny parts. Encapsulation creates a new data type, an abstract data type. Thinking in terms of objects is
like thinking about music in terms of chords instead of individual notes. This sounds great, but rarely have I
encountered worth-while problems that benefited from an abstract data type. But then again, I stick to small
school problems. What I have encountered are artificial problems designed to require encapsulation for student
learning.

   However encapsulation is design, and an efficient design often comes from throwing out several inefficient
designs. You can spend a lot of time trying to produce a generic class. The experts give us this advice:

   a.	Try to write natural functions that closely correspond to reality. The whole point of classes
        (abstractions) is that they should make thinking and programming more intuitive.

   b.	Rather than trying to design a near-optimal class, design it so that it is easy to extend.




      Despite many claims that promise a smooth transition from object-oriented analysis to design, in
      practice the transition is anything but smooth. A flexible and reusable design will contain
      objects that aren't in the analysis model. The programming language and class libraries you use
      affect the design. Analysis models often must be re-designed to make them reusable.--Erich Gamma,
      Richard Helm, Ralph Johnson, and John Vlissides (the "gang of four"), Design Patterns, Elements
      of Reusable Object-Oriented Software (Addison-Wesley, 1995), pages 11, and 353.

   With encapsulation comes data hiding with private data--.e., data that either cannot be accessed or can
only be accessed through getters and setters, which limit modifications. The user could, of course, rewrite
the class so that the private data could be accessed with no limits, but then this would be a different
class. The advantage of private data is that the class can be trusted to do what it claims to do. If a
class has been well-tested, the bugs in your program are unlikely to be found in the class you imported.
Of course, the same can be said about the tested functions in any library module.

   I once wrote a neural network program. Initially I thought making a Node class made sense, because a
network is made up of nodes. But the Node class seemed to complicate matters. So I rewrote the network without
any classes. Then I decided to re-write the network program again as a single neural-network class. This
shouldn't make any sense, because I would have only a single object. What's the point? But, in fact, it made
the coding simpler, because of the semi-global properties of the many internal class parameters: I didn't have
to pass or return them. Since the network was small, the negative side-effects of globalization didn't occur.

   Polymorphism is possible without classes, but polymorphism can be especially effective with operator
overloading. For example, when working with vectors, we can overload all of the operators in a vector class
and end up writing this

                                            F = 3*(B+C)/4 - A/2,
instead of this:

                      F = vectMinus(scalarMult(3/4, VecAdd(B,C)), scalarMult(1/2, A)).

   Now you see why I built a vector class. The class overloading really did make a positive difference. But I
only needed it for about ten lines of code. Was it really worth the effort? Java allows programmers to
overload functions, but not operators. Python does the reverse: You can overload Python's operators, but not
its functions. One reason function overloading is not needed in Python is that all the parameters sent to a
function can be packed into a list and sent as one parameter. Or they can be sent separately and, using the
star operator (*), received as a single list.

   In both C++ and Python you can overload operators that already exist, but cannot introduce new operators.
Continuing with our vector example, if I want to write a line of code involving the cross product, I can NOT
use "x" as an operator. Instead I must write something like A = B.crossProd(C), or A = Vector.crossProd(B,C).
I suppose the reason is that "x" is a text letter, and therefore cannot be an operator. But not being able to
use an imported symbol as an operator seems like a language deficiency to me. Below is a Zen-like quote from
the Father of C++, the man who put classes in to C.

      If you can think of "it" as a separate idea, make it a class. If you can think of "it" as a
      separate entity, make it an object of some class.--Bjarne Stroustrup, The C++ Programming
      Language, 2nd ed. (Addison Wesley, 1991), page 10.

   What does it mean? Dr. Stroustrup didn't give any examples. But I don't think it means much for small
school programs. Classes make sense in huge programs where code can be written in terms of the objects. But in
most school problems, which are usually implemented algorithms, a student programmer needs a compelling reason
to introduce a class into his code.

                                             The End
