"""+=================+=================*=================+=================+
   ||                                                                     ||
   ||             IF, ELIF, ELSE, TERNARY IF-ELSE, and BOOLEAN            ||
   ||                                                                     ||
   +=================*===================================*=================+

   Examples:
              #---standard
                  if a < 4:
                     print (a)
                  else:
                    print (b)

              #---ternary
                  print (a) if a < b else print(b) # no colons and no commas

              #---better ternary
                  print (a if a < b else b)        # no colons

              #---weird ternary (Don't write code like this.)
                  a = 3; b = 4
                  print([a, b][a <= b])

              #---if, elif, and else
                  if a > b:
                    print ('>')
                  elif a == b:
                    print ('=')
                  else:
                    print ('<')

              #---weird if, elif, and else (Don't write code like this.)
                  a = 5; b = 4
                  print(  (a<b and 'less') or (a==b and 'equal') or (a>b and 'greater') )


                # RUN time error (beginner's error 1, bleeding ifs)
                    x = 1
                    if x == 1: x = 2
                    if x == 2: x = 3
                    if x == 3: x = 4
                    print (x) # output: 4

                # RUN time error (beginner's error 2, back stabbing else)
                    x = 1
                    if x == 1: x = 2
                    if x == 3: x = 4
                    else:      x = 5
                    print (x) # output: 5

                # LOGIC error (beginner's error 3, misaligned else. output should have been 5)
                    x = 2; y = 3
                    if x == 1:
                        x = 2
                        if  y == 3:
                            z = 4
                        else:
                            x = 5
                    print (x) # output: 2

                #  logical without else and elif. The code is easy to debug.
                    x = 1
                   if x == 1:
                      x = 2
                      return # output: x = 2 is returned
                   if x == 2:
                       x = 3
                       return
                    if x == 3:
                       x = 4
                       return

                # The else and elif improve the previous code.
                    x = 1
                    if   x == 1: x = 2
                    elif x == 2: x = 3
                    elif x == 3: x = 4
                    else: x = 5
                    print (x) # output: 2


   Don't forget DeMorgan's laws used for simplification:

                    not (A and B) = (not A( or  (not B)
                    not (A or  B) = (not A) and (not B)

   Here a way check if you have an int type that is greater than or equal to 1. The second case reduces to the
first case by use of DeMorgan's Laws.

           if type(num) != int or num < 1:
               exit('Bad data 1')
           if not (type (num) is int and num >= 1):
               exit('Bad data 3')

   If statements, like loops, can compact their bodies with semicolons. The following examples are all
equivalent.

                  if a > b: print(3); print(4) # avoid

                  if a > b:
                     print(3); print(4)        # avoid

                  if a > b:                    # preferred
                     print(3)
                     print(4)

   There is no switch statement and no case statement in Python. The term "elif" stands for "else if". All
numbers are True, except 0, which is False. The boolean True is mapped to 1 and False is mapped to 0.

   Nevertheless, you can construct a (anonymous) switch-type statement with a dictionary. Below, if choice =
1234, then Smith will be printed.

   print({1234: 'Smith',
          6834: 'Jones',
          3987: 'Kane',
          6835: 'Orwell',
          }[choice])

    An empty set, empty list, empty tuple, empty dictionary, empty string, and the value None are all
evaluated as the boolean False. The advantage of this policy is that you can test an object directly, instead
of comparing it to another object. Do not write "if L == []: ...". Instead, write "if L: ...". And in
general, avoid "if a == true: ...", and avoid "if a != 0: ...". Instead write "if a: ...".

   Compound expressions are really and statements--e.g.,

               a < b <= c is intrepreted as (a < b) and (b <= c)

   Be careful about chaining.

    print (0<1 == True ) # Output: True:  (0<1) and (1==1)   --> True
    print (1<0 == False) # Output: False: (1<0) and (1==1)   --> False
    print (2<3 == True ) # Output: False: (2<3) and (3 == 1) --> False
    print (2<3 == False) # Output: False: (2<3) and (3 == 0) --> False

   Note that not(0 < n < 10) is equivalent to 0 >= n >= 10, which I think is more clearly written unchained as

                    n <= 0 or 10 <= n (larger on right side, smaller on left side).

   Python uses SHORT-CIRCUIT evaluations--e.g. division by zero cannot occur in the following expression.

                        if x !=0 or 7/x > 1: ...

   Why? Because if the first part of an OR statement is true, then the second part needs not be evaluated. If
the first part of an AND statement is false, the second part is not evaluated. This is not only useful, but
necessary to understand Python boolean AND and OR statements. AND and OR statements do not return True, False,
1 or 0. Instead, they return the last object they evaluated.

                    print (3 and 4)       # output: 4
                    print (3 or  4)       # output: 3
                    print (not 4)         # output: False
                    print (3 and (not 4)) # output: False
                    print (1 != 4)        # output: True
                    print (int(True))     # output: 1

   All numbers are True, except 0, which is False. The boolean False is equivalent to 0. The boolean True is
equivalent to 1. Thus, False < True.

  Consider this line:

                    x = a or b or c or d or None

   The variable x will take on the first variable that is true, or else x will take the value None.

   a = 5; b = 4
   print(a < b and 'less' or a==b and 'equal' or a > b and 'greater')
                                  # output: greater

   Above, the ands are done first.


   If comparisons get complicated, then try keeping your comparisons in number-line order--e.g., do not write
"if x > 10 and x <= 12:...", but rather "if x <= 12 and 10 < x".

   Complicated tests should be turned into boolean function calls. That is, place the test in a function and
return True or False.

   One of the great traps in programming with if-statements is the mistake of using > instead of >=, or vice
versa. This, of course, comes from not thinking carfully about the problems we are solving. Below are examples
of boolean functions.


   from random import random
   print( int(random() > 0.25))   # prints 1 75% of the time and prints 0 25%.

   print ('1.', '1234'.isdigit()) # output: True

   print ('2.', '1234'.isalpha()) # output: False
   print ('3.', 'abCD'.isalpha()) # output: True
   print ('4.', 'ab D'.isalpha()) # output: False (The space makes it false.)

   print ('5.', '1234'.isalnum()) # output: True alnum = alpha-numeric
   print ('6.', '1aBC'.isalnum()) # output: True
   print ('7.', '-1.0'.isalnum()) # output: False
   print ('8.', '12 3'.isalnum()) # output: False (The space makes it false.)

   print ('9.', 'abcd'.islower()) # output: True
   print ('0.', 'abcd'.isupper()) # output: False

   a = 1; s = 'ABC'; f = 1.2; L = [1,2,3,]
   print( type(a) == int)     # output: True
   print( type(s) == str)     # output: True
   print( type(f) == float)   # output: True
   print( type(L) == list)    # output: True

   print(isinstance(a, int))  # output: True
   print(isinstance(s, str))  # output: True
   print(isinstance(f, float))# output: True
   print(isinstance(L, list)) # output: True

   print(type(a) == float)    # output: False
   print(isinstance(a, float))# output: False
