"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                                     NUMBERS and OPERATORS                                   ||
      ||                                                                                             ||
      +=================-=========-=========-=========*=========-=========-=========-=================+


   NUMBERS and BOOLEANS. Python has no unsigned types.
      1. int     (unlimited size, restricted only by memory),
      2. float   (16-17 significant digits),
      3. Decimal super-precision floats and ints, with some limitations.
      4. complex (a pair of floats),
      5. boolean (True and False, mapped to 1 and 0, respectively). All non-zero
         numbers are considered True.
      6. None    This is the default value returned by all functions. None is similar to null in Java.
      7. The expression float('inf') is greater than all numbers, and float('-inf') is less than all numbers
         --e.g.,

      Examples:
        if 3 < float('inf'): print ('less') # output: less
        print (2 > 1)                       # output: True
        print(bool(2), bool (0))            # output: True False
        print (int(True), int(False))       # output: 1 0

   HEX and BINARY NUMBERS.

                print(bin(7))      # output: 0b111
                print(bin(7)[2:])  # output: 111
                print(hex(17))     # output: 0x11
                print(hex(17)[2:]) # output: 11
                print (int(str(32),  8)) # output: 26
                print (int(str(11), 16)) # output: 17
                print (int(float.fromhex('12'))) # output: 18

   FLOATS. Look at this:

            print (0.1 + 0.1       == 0.2)   # output: True
            print (0.1 + 0.1 + 0.1 == 0.3)   # output: False
            print((0.1 + 0.1 + 0.1) - 0.3 )  # = 5.551115123125783e-17

    Evidently if there is ANY difference between two floats, then they are not considered equal. This is too
    sensitive of a difference for most applications.  As a practical consideration, if the difference between
    two floats is less than a trillionth,

                                       1e-12 = 0.000 000 000 001,

    I claim the two float numbers are equal. What is the basis for the one trillionth tolerance of error? If
    we continue to add 0.1 to a variable, then after 929 additions the total is 92.899999999999, which is
    slightly more than one-trillionth in error. The results for other error tolerances are given below.
    Intuitively, I prefer the nearly-1000-repetitions-of-a-tenth before an error occurs.

               n =  262  t =  26.200000000000102, for 1e-13
               n =  929  t =  92.899999999999,    for 1e-12
               n =  2513 t =  251.29999999999,    for 1e-11

    Moral: Floating point numbers are NOT real numbers. The number 0.1 can NOT exactly be represented in
    binary, only approximated. Here is a function that will return floating-point equality for you. Place it
    in your personal library.

                 +----------------------------------------+
                 |   def equal(num1, num2):               |
                 |      return abs(num1 - num2) < 1e-12:  |
                 +----------------------------------------+

     By the way, "is" is for memory addresses, and "==" is for values. Rarely do you want to use "is".

     Python 3.x has unlimited INTEGER precision (yea!). See the section on the DECIMAL type.
#-------------------------------------------------------------------------------------------------------------


  46. TYPE CASTING. Type casting is common in Python: int(x), float(x), str(x), list(x),
      and tuple(x). Below we cast an integer to a string, to a list, back to a string,
      and then back to an integer:

                      num = 1234     # = 1234
                      s = str(num)   # = '1234'
                      L = list(s)    # = ['1', '2', '3', '4']
                      s = ''.join(L) # = '1234'
                      num = int(s)   # = 1234

   OPERATORS. Python has two symbols for divide (/ and //):

          // = divide and round down (not truncate)

          examples: 7//2 =  3, an int type.
                   -7//2 = -4, an int type.
                    7/2  = 3.5, a float type.
                    8/2  = 4.0, a float type.

      In Python 2.x the symbols '/' and '//' are reversed in function.

   Python has both an exponent operator (**), and an exponent function (pow):

            from math import pow
            print(pow (64, 1/3)) # output:  3.9999999999999996
            print(64 ** (1/3))   # output:  3.9999999999999996
            print(pow (4, 3))    # output: 64.0, float type
            print (4 ** 3)       # output: 64,   int   type

      Note the difference between the ** operator and the pow function:

      print(2**1023) # There is no limit, if base and exponent are both ints.
      # output: 89884656743115795386465259539451236680898848947115328636715040578866337902
      #         75048156635423866120376801056005693993569667882939488440720831124642371531
      #         97370621888839467124327426381511098006230470597265414760425028844190753411
      #         71231440736956555270413618581675255342293149119973622969239858152417678164
      #         812112068608

      print(pow(2, 1023)) # But pow(2, 1024) = OverflowError: math range error.
      # output: 8.98846567431158e+307

      print(2.1**956) # But 2.1**957 = OverflowError: (34, 'Result too large')
      # output: 1.100641200053886e+308

   The Python arithmetic operators are +, -, *, /, //,  ** (power),  % (mod), + (unary), and - (unary). Here
is a tiny code fragment to print prime numbers:

                +-------------------------------------+
                |   p = [2,]                          |
                |   for n in range (3, 100):          |
                |       for divisor in p:             |
                |           if n%divisor == 0: break  |
                |       if n%divisor == 0: continue   |
                |       p.append(n)                   |
                |       print (n, end = ", ")         |
                +-------------------------------------+
#-------------------------------------------------------------------------------------------------------------
   ORDER OF OPERATIONS (aka operator precedence, hierarchy of operations). Whenever an expression is followed
and preceded by operators, the operator with the highest precedence is applied first. If the precedences are
equal, then the operator that precedes the expression is first applied. For example,

                       x = 3 <= 4 == 1
                       print (' 2.', x) # output: False.

   Thus, you might be tempted to think that "==" comes before "<=". But this is not true. Any chained ex-
pression is translated into an and-statement:

              3 <= 4 == 1 is translated into (3 <= 4) and (4 == 1).

   Below is a partial order of operations (mnemonic FPUARLA: Fine Push-Ups Are Really Lousy Activities).
Perhaps you can do better. Note that functions and subscripting come before (bind more tightly) all operators
below, where they can be combined. BTW, the asterisk (*) is often called "star" operator, the "splat" (a small
joke), and the Nathan hale (a big joke). "I regret that I have but one ...".

          0. Functions and subscripting
          1. Power: **
          2. Unary: +,  -
          3. Arithmetic: *, /, //, % (mod)
          4.            +, - (Multiplication, division, and mod are done before add and subtract.)
          5. Relations <, <=, >, >=,  ==, != [Yes, all on the same line.]
          6. Logic: not
          7.        and ("not" is done before "and".)
          8.        or  (Both "not" and "and" are done before "or".)
          9. Assignments: ==, +=, -=, *=, /=, %=  <-- assignments are done last
             There is no ++ operator in Python.

   Below is a working program.

      Examples:
         print(-2**2)     # = -4
         print(2 + 3 * 4) # = 14
         print(not True and False)      # = False ("not" was done first)
         print(True or False and False) # = True  ("and" was done first)

         x = 2*[0, 1, 2,]
         print (x) # output: [0, 1, 2, 0, 1, 2]
         x = 2*[0, 1, 2,][2]
         print (x) # output: 4, because subscripting comes before multiplication.

         x = 2
         print (1/2*x) # output: 1.0

         a = False
         b = True
         if (a and b != True):
            print(True)
         else:
            print (False)  # output: False, because "!=" comes before "and".

         x = not [0, 1, 2,]
         print (x) # output: False

         x = not [0, 1, 2,][0]
         print (x) # output: True, because subscripting comes before "not".

         How do we interpret the expression a < b == c < d? The answer is a < (b == c and c < d).
         But who wants to remember non-intuitive details like this? Instead use parentheses. The person
         writing a < b == c < d probably means (a < b) == (c < d).

         Moral: Keep chained expressions simple: a < b < c obviously means a < b and b < c. No one is
         confused. But a < b == c < d needs parentheses.

"""
#-------------------------------------------------------------------------------

from math import log10, log, sqrt, trunc, ceil, floor, exp, sin, cos, fabs, \
                 tan, asin, acos, atan, atan2, degrees, radians, hypot, \
                 factorial, pi, e, fabs
#-------------------------------------------------------------------------------
from math    import * # sqrt, log10, ln, exp, etc.
from decimal import * # Decimal and getcontext().prec




def main():
    a = 1; b = 2; c = 0
    bool1 = (a != b < c)
    bool2 = (a != b and b < c)
    print(bool1)
    print(bool2)
    print (factorial (5))# output: 120
    exit()

    print (2.5.as_integer_ratio()) # output: (5, 2), meaning the fraction 5/2.

    print (abs (-37.3)) # output: 37.3
    print (fabs(-37.3)) # output: 37.3, fabs must be imported.
    print (abs (-37))   # output: 37
    print (fabs(-37))   # output: 37.0

    print (float('inf'))      # greater than all numbers
    print (float('-inf'))     # less    than all numbers
    print (float('infinity')) # same as float('inf')
    print (float('nan'))      # not a number

    print (pi)           # output: 3.141592653589793 (imported from math)
    print (e)            # output: 2.718281828459045 (imported from math)

    a = 1234567890123456789
    b = 9876543210987654321
    print (a*b)          # output: 12193263113702179522374638011112635269

    print (divmod(7,3))  # output: (2,1)
    print ( 8 // 3)      # output:  2, truncates with positive numbers
    print (-8 // 3)      # output: -3, rounds DOWN with negative numbers
    print (8.0 // 3)     # output:  2.0,a float
    print (7 % 3)        # output:  1

    print (pow(8, 1/3))  # output: 2.0
    print (8 ** (1/3))   # output: 2.0
    print (8 **  1/3 )   # output: 2.6666666666666665. Beware missing parentheses.

    print (sqrt(25))     # output: 5.0
    print (hypot(3, 4))  # output: 5.0 = sqrt(3 * 3 + 4 * 4)
    print (sum((1,2,3))) # output: 6

    print (log (2))      # 0.6931471805599453. float types are 16-17 digits.
    print (log(2.718281828459045)) # output: 1.0
    print (log10(1000))  # output: 3.0
    print (exp(1))       # output: 2.718281828459045. exp is inverse of log()

    print (max(1,2,3))   # output: 3
    print (min(1,2,3))   # output: 1
    print (factorial (5))# output: 120

    print (bool(-1.23))  # output: True
    print (bool(0))      # output: False

    print (sin(1.2))     # output:0.9320390859672263
    print (cos(1.2))     # output:0.3623577544766736
    print (tan(1.2))     # output:2.5721516221263188

    print (asin(0.9))    # output: 1.1197695149986342
    print (acos(0.9))    # output: 0.45102681179626236
    print (atan(1))      # output: 0.7853981633974483 = 45 degrees
    print (atan2(3,2))   # returns -pi < atan (3/2) < pi

    print (degrees(1))   # output: 57.29577951308232
    print (radians(57))  # output: 0.9948376736367679
    print (pi)           # output: 3.141592653589793
    print (e)            # output: 2.718281828459045

    x = 37.086
    print (round( x))    # output:  37
    print (round(-x))    # output: -37

    print (round( x, 2)) # output:  37.09
    print (round(-x, 2)) # output: -37.09

    print (trunc( x))    # output:  37
    print (trunc(-x))    # output: -37

    print (int( x))      # output:  37
    print (int(-x))      # output: -37

    print (ceil ( 2.4))  # output:  3
    print (ceil (-2.4))  # output: -2


    print (floor( 2.4))  # output:  2
    print (floor(-2.4))  # output: -3

    print (ord('A'))     # output 65
    print (chr(65))      # output 'A'

    print (bin(9))       # output: 0b1001
    print (0b1001)       # output: 9
    print (int('1001',2))# output 9

    print (oct(26))      # output: 0o32
    print (0o32)         # output: 26
    print (int(str(32), 8))# output: 26
    print ('---------------------------')
    print (hex(26))      # output: 0x1a
    print (0x1a)         # output: 26
    print (int(str(11), 16))# output: 17
    print (float.fromhex('12')) # 18.0 (works only with floats)

    a = 1; b = 2
    print(complex(a, b)) # output: 1 + 2j
    z = 2 + 3j           # Note: j is a special operator.
    print(z)             # output: 2+3j
    print(z.conjugate()) # output: 2-3j
    print(z.real, z.imag)# output: 2.0 3.0 (both are floats)
    import cmath
    print (cmath.sqrt(-1))        #output: 1j
    print (cmath.polar(sqrt(2)+1.4142135623730951j)) # (2.0, 0.7853981633...)
    print (cmath.rect (2, radians(45))) # output: 1.414+1.414j
    x = 9
    print (isinstance(x, int))       # output: True
    print (isinstance(x, float))     # output: False
    print (isinstance(x,str))        # output: False
    print (type(x)   == int)         # output: True
    print (type(3.0) == float)       # output: True
    print (type([x]) == list)        # output: True
    exit()
    print (list(range( 5)))          # [0, 1, 2, 3, 4]
    print (list(range(-5)))          # []
    print (list(range(0, 5, 2)))     # [0, 2, 4] (3 parameters are required)

    print (0.375.as_integer_ratio()) # output: (3, 8) = 3/8 (floats only)
    print (3.0.is_integer())         # output: True (works with floats only)
#-------------------------------------------------------------------------------
if __name__ == '__main__': main()
#=============================== END OF SECTION ===============================#
