"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                                            SETS                                             ||
      ||                                                                                             ||
      +=================-=========-=========-=========*=========-=========-=========-=================+


   A set is an UNORDERED collection of UNIQUE elements. Like lists and tuples, the elements of a set can be of
different types. Sets can contain only immutable elements, but sets themselves are mutable. If you cast a list
to a set, duplicates will automatically be deleted (handy). You cannot index a set. You cannot slice a set.
Checking membership in a set of n elements is constant time--i.e., the time is not dependent on the number of
items in the set. A related type is the frozenset, which is an immutable set:

                                          s = frozenset({1,2,3}).

   Since sets are mutable, they cannot be stored in other sets, unless they are of the frozen set variety. Set
comprehensions are allowed:

                                    s = {n**2+1 for n in range(3)}
                                    print (s) # output: {1, 2, 5}

   Sets are useful for determining membership, unions, intersections, and filtering out duplicates (cast to
set type, and cast back again).

   You APPEND elements to a set (here called s) with
      a) the s.add(3) function (for a single element) and with
      b) the s.update({4,5}) function (for a set of elements).

   You REMOVE elements from a set with
      a) s.remove (3). Causes an error if element is missing.
      b) s.discard(4). Does NOT cause an error if an element is missing.
      c) s.pop. Removes a random element from the set. Set must NOT be empty.

   WARNING: {} is NOT an empty set; it is an empty dictionary. This is an empty set: set().

   The set symbol operators are:
    1. - (difference),
    2. | (union),
    3. & (intersection),
    4. ^ (symmetric difference).

    These operators work only with two set-types. However, their functional counterparts will work with a set
and a tuple, and a set and a list.

                   s = {1,2}
                   t = (1,2,3)
                   l = [1,2,3,4]
                   print(s.union(t)) # output: {1, 2, 3},    even though t is NOT a set.
                   print(s.union(l)) # output: {1, 2, 3, 4}, even though l is NOT a set.
                   print(l.union(s)) # ERROR, because l is NOT a set.

   Here are some examples:

   s = {}                         # an empty dictionary, NOT a set. So usually this is a mistake for sets.
   s = set()                      # the null set
   s = {1,2,}
   s.clear()
   print ('s =', s)               # s = set()

   s = {1, 2, 3, 2}
   print ('s =', s)               # output: s = {1, 2, 3}

   s = set([1, 2, 3, 3])
   print ('s =', s)               # output: s = {1, 2, 3}

   s = set('ABBCEED')
   print ('s =', s)               # output: s = {'A', 'C', 'B', 'E', 'D'}

   print (len(s))                 # output 5

   print (2   in s,  2  not in s)     # output: False rue

   print ('B' in s, 'B' not in s)     # output: True False

   s = {1,2,}
   s.add(3)                       # add() is for single element only
   s.update({4,5})                # update() is for a set of elements
   print (s)                      # output: {1, 2, 3, 4, 5}

   # s.remove(7)                  # ERROR: item 7 is not in set.

   s.discard(7)                   # no error, but 7 is not in set, either

   s.discard(4)
   print ('s =', s)               # output: s = {1, 2, 3}

   s = {1,2,3}
   x = s.pop()
   print ('s =', s, 'and x =', x) # output: s = {2, 3} and x = 1
                                  # removed and returned a RANDOM element
                                  # from set s
   s1 = {'A', 'B', 'C'}
   s2 = {'C', 'D', 'E'}

   print ('s =', s1-s2)           # output: s = {'A', 'B'}. The 'D' and 'E' are ignored.

#  symmetric difference is what is NOT in both sets, the opposite of union.
   print ('s =', s1.symmetric_difference(s2)) # output: s = {'A', 'B', 'E', 'D'}
   print ('s =', s1 ^ s2)                     # output: s = {'A', 'B', 'E', 'D'}.

   print ('s =', s1.union(s2))    # output: s = {'A', 'C', 'B', 'E', 'D'}
   print ('s =', s1|s2)           # output: s = {'A', 'C', 'B', 'E', 'D'}

   print ('s =', s1.intersection(s2)) # output: s = {'C'}
   print ('s =', s1 & s2)             # output: s = {'C'}

   print (s1.issubset  (s2))      # False
   print (s1 < s2)                # False
   print (s1.issuperset(s2))      # False
   print (s1 > s2)                # False
   print (s1.isdisjoint(s2))      # False

   #print (s[0])                  # ERROR, no indexing allowed with sets.
   print(set("Hello, mello?"))
   # output from above: {' ', 'e', 'H', 'm', 'l', 'o', ',', '?'}
   print(set(["Hello", "mello?"]))         # output: {'mello?', 'Hello'}
   print(set("".join(["Hello", "mello?"])))#  {'e', 'H', 'm', 'l', 'o', '?'}

   The look-up speed for a list is LINEAR time. But the look-up speed for a set is CONSTANT time. In some
situations it may make sense to cast a list to a set. Here is a set trick to limit a range:

   Lst = [1,3,]

   for n in set(range(8)) - set(Lst):
       print(n, end = ', ')    # output: 0, 2, 4, 5, 6, 7,
   print()

   The above can also be accomplished like this:
   for n in range(8):
       if n not in Lst:
          print(n, end = ', ') # output: 0, 2, 4, 5, 6, 7,
