""" +===============================+========-========*========-========+===============================+
    ||                               UNICODE, UTF-8, ASCII, Latin-1, etc.                              ||
    +===============================+========-========*========-========+===============================+


   To understand Unicode, you need to understand some history. So, here we go.

   ASCII (American Standard Code for Information Interchange) is a character format scheme (associating
characters with numbers) that associates 128 common English language symbols with 7-bit binary numbers. (The
7-bits are always embedded in 8-bit bytes.) The scheme was introduced in 1960 and was first published as a
standard in 1963. The first 32 characters (binary numbers 0-31) and character 127 are teletype control
characters, which are now mostly obsolete. Character 31 is the blank space, characters 48-57 are the digits
0-9, uppercase characters A-Z are 65-90, and lowercase characters a-z are 97-122. An Extended set of ASCII
codes were used for a while on some computers. These were mostly box making symbols, mathematical symbols, and
certain accented characters from non-English languages.


                 32 = space 33 = !     34 = "     35 = #     36 = $     37 = %     38 = &
                 39 = '     40 = (     41 = )     42 = *     43 = +     44 = ,     45 = -
                 46 = .     47 = /     48 = 0     49 = 1     50 = 2     51 = 3     52 = 4
                 53 = 5     54 = 6     55 = 7     56 = 8     57 = 9     58 = :     59 = ;
                 60 = <     61 = =     62 = >     63 = ?     64 = @     65 = A     66 = B
                 67 = C     68 = D     69 = E     70 = F     71 = G     72 = H     73 = I
                 74 = J     75 = K     76 = L     77 = M     78 = N     79 = O     80 = P
                 81 = Q     82 = R     83 = S     84 = T     85 = U     86 = V     87 = W
                 88 = X     89 = Y     90 = Z     91 = [     92 = \     93 = ]     94 = ^
                 95 = _     96 = `     97 = a     98 = b     99 = c    100 = d    101 = e
                102 = f    103 = g    104 = h    105 = i    106 = j    107 = k    108 = l
                109 = m    110 = n    111 = o    112 = p    113 = q    114 = r    115 = s
                116 = t    117 = u    118 = v    119 = w    120 = x    121 = y    122 = z
                123 = {    124 = |    125 = }    126 = ~    127 =     128 = none


   EBCDIC (Extended Binary Decimal Interchange Code) is an 8-bit format scheme used on the highly successful
IMB mainframe computers. The scheme was an extension of the scheme originally used by IBM business machines
that used punch cards in the 1960s. Two disadvantages of the EBCDIC scheme was that different variants existed
and letter placements were not contiguous--add 1 to the number for 'A' will not get the number for 'B'.
Curiously, the first IBM PC (personal computer) used the ASCII scheme, not their own EBCDIC scheme.

   LATIN-1 is an 8-bit scheme that is the same as ASCII for the first 128 (0-127) bits, but then allows
another 128 characters for ordinal numbers 128-255.  On some lists, the numbers 128-159 are not associated
with any characters. Latin-1. was popular in Europe and is currently being phased out in favor of UTF-8, the
"king of encodings".

   The UNIVERSAL CHARACTER SET is a HUGE set of characters that is currently growing and that its creators
hope will eventually contain every known symbol to mankind. At this time, the UCS has more than 1.1 million
code points (numbers) representing specific symbols: letters, digits, ideograms, symbols from ancient writing
systems, and even symbols from fantasy writing systems. The UCS currently leaves many in-between-numbers
unassigned to accommodate the possibility of symbols to be added later. Currently, only about 10% of the
possible code points have actually been assigned characters, and that includes almost all of the major living
languages.

   The UCS is divided into groups called planes, each plane containing 65,536 (= 2**16) characters. The BASIC
MULTILINGUAL PLANE is the first set of characters and was released in 2000. Unicode has several formatting
schemes. UTF-8 (Unicode Transformation Format, designed by Ken Thompson and Rob Pike) has become the dominant
character encoding scheme for the World Wide Web, a kind of universal text font. UTF-8 encodes each of the
1,112,064 code points (numbers associated with characters in the Unicode character set using one to four 8-bit
bytes).

   Character codes from    0 to    127 are represented by 1 byte  (ASCII).
   Character codes from    0 to    255 are represented by 1 byte (Latin-1).
   Character codes from  256 to   2047 are represented by 2 bytes
                   from 2048 and above are represented by 3 or 4 bytes (UTF-8).

   Note: ASCII is a subset of Latin-1 which is a subset of UTF-8. In fact the code numbers are the same in all
three sets. All three are considered Unicode.  Recall ord('A') = 65 = 0x41, ord('B') = 66 = 0x42, and ord('C')
= 67 = 0x43 (the prefix "0x" means hexadecimal.) Below are three ways to print a Unicode character.

     print('\x41\x43')                   # output: AC  (8-bit hex     escapes)
     print('\u0041\u0043')               # output: AC (16-bit Unicode escapes)
     print(chr(65), chr(0x43), sep = '') # output: AC (the character function)

   So what does our computer's character set look like? The answer depends on whether your output is going
to the console window, to a printer, or to a graphics screen. In graphics mode all characters seem to be
printed. [I will give you a tiny program at the end of this discussion that will do this for you.] In the
console window and on the printer some characters are printed as a space (" "), as nothing (""), or as a tiny
box. Nevertheless, the printer and console window can print a good part of the Unicode set.

  So what do we see in UTF-8? The numbers 0 to 31 are printing commands (non-printable characters). Character
32 is the blank space, which is considered a printable character. Character 160 is the NON-breaking-blank-
space. Character 127 is also an unprintable character (actually a teletype/printer command). Characters 128-
159 are left blank for some reason. After that there are groups of characters and groups of unassigned char-
acters. I couldn't find any characters after 65,518. In fact, the last Unicode ordinal my laptop will even
accept is 65,535 = 0xFFFF. A number bigger than 65,518 returns this error: "character U+10000 is above the
range (U+0000-U+FFFF) allowed by Tcl".

   In the console window, sometimes the printing of one character can change the font of the characters that
follow. Sometimes printing of one character can change the looks, or even suppress the character that
follows--e.g., print (chr(128) + chr(65533)) and print (chr(160) + chr(65533)).

   What is past character 255? See the Internet for a list. Box drawing characters are 9472-9621.

        chr(9679)  = solid disk.        chr(9731 = snowman
        chr(9829)  = heart.             chr(65533) = diamond with a small hole near the top.

    Below is a tiny program that will print characters into a graphics window. I never got around to allowing
the font size to be easily changed from the default 15, or making the program do paging. Here is a tiny
project that needs to be finished.
"""
##############################################<START OF PROGRAM>##############################################
def setUpCanvas(root):
    root.title("A Tk/Python Graphics Program")
    canvas = Canvas(root, width = 1270, height = 780, bg = 'DARK BLUE')
    canvas.pack(expand = YES, fill = BOTH)
    return canvas
#-------------------------------------------------------------------------------------------------------------

def printUnicode (startNumber = 0):
    row  = 22  # number of rows
    col  = 10  # number of columns
    size = 15  # font size of char. Unfortunately, this is fixed at 15. Increase and characters will overlap.
    canvas.create_text(125, 20, text = "UTF-8 (starting at " + str(startNumber) +")", fill = 'WHITE', \
                       font = ('Helvetica', 12, 'bold') )
    for r in range(row):
        for c in range(col):
            num =c+r*col+startNumber
            tab = ' '* (col-len(str(num)) + 5)
            canvas.create_text(c*120 + 48, r*25 + 45, text = tab+str(num) + " = "+ \
                   chr(num), fill = 'GOLD', font = ('Helvetica', size, 'bold') )

#---------------------------------------<GLOBAL CONSTANTS AND IMPORTS>----------------------------------------
from tkinter   import * # Graphics (capitalized as Tkinter in Python 2.x)
root=Tk()
canvas = setUpCanvas(root)
#===================================================<MAIN>====================================================

def main():
    printUnicode (9728)
    root.mainloop()
#-------------------------------------------------------------------------------------------------------------
if __name__ == '__main__':  main()
###############################################<END OF PROGRAM>###############################################
"""